#include <catch.hpp>
#include "../common.h"

#include <pfor/expression/tagger.h>
#include <pfor/expression/operators.h>

TEST_CASE("ExpressionTagger") {
	constexpr pfor::expr::Access R = pfor::expr::Access::read, W = pfor::expr::Access::write;

	using O1 = pfor::expr::Expression<int*, class IDa, pfor::Index>;
	using O2 = pfor::expr::Expression<int*, class IDb, pfor::Index>;
	using O3 = pfor::expr::Expression<int*, class IDc, pfor::Index>;

	using OT1R = pfor::expr::OperandTag<class IDa, R, pfor::Index>;
	using OT2R = pfor::expr::OperandTag<class IDb, R, pfor::Index>;
	using OT3R = pfor::expr::OperandTag<class IDc, R, pfor::Index>;
	using OT1W = pfor::expr::OperandTag<class IDa, W, pfor::Index>;
	using OT2W = pfor::expr::OperandTag<class IDb, W, pfor::Index>;
	using OT3W = pfor::expr::OperandTag<class IDc, W, pfor::Index>;

	using Tags1R       = pfor::Pack<OT1R>;
	using Tags1R2R     = pfor::Pack<OT1R, OT2R>;
	using Tags1R2R3R   = pfor::Pack<OT1R, OT2R, OT3R>;
	using Tags1W2R     = pfor::Pack<OT1W, OT2R>;
	using Tags1W1R2R   = pfor::Pack<OT1W, OT1R, OT2R>;
	using Tags1W1R2R3R = pfor::Pack<OT1W, OT1R, OT2R, OT3R>;
	using Tags1W2W3R3R = pfor::Pack<OT1W, OT2W, OT3R, OT3R>;
	using Tags1R2W3W3R = pfor::Pack<OT1R, OT2W, OT3W, OT3R>;
	using Tags1R2W3R3R = pfor::Pack<OT1R, OT2W, OT3R, OT3R>;

	using EPlus           = decltype(+V(O1));
	using EMinus          = decltype(-V(O1));
	using ENot            = decltype(!V(O1));
	using EBitNot         = decltype(~V(O1));
	using EIndirection    = decltype(*V(O1));
	using EAddressOf      = decltype(&V(O1));

	using EMultiplication = decltype(V(O1) * V(O2));
	using EDivision       = decltype(V(O1) / V(O2));
	using EModulo         = decltype(V(O1) % V(O2));
	using EAddition       = decltype(V(O1) + V(O2));
	using ESubtraction    = decltype(V(O1) - V(O2));
	using ELeftShift      = decltype(V(O1) << V(O2));
	using ERightShift     = decltype(V(O1) >> V(O2));
	using ELower          = decltype(V(O1) < V(O2));
	using ELowerEqual     = decltype(V(O1) <= V(O2));
	using EGreater        = decltype(V(O1) > V(O2));
	using EGreaterEqual   = decltype(V(O1) >= V(O2));
	using EEqual          = decltype(V(O1) == V(O2));
	using ENotEqual       = decltype(V(O1) != V(O2));
	using EBitAnd         = decltype(V(O1) & V(O2));
	using EBitXor         = decltype(V(O1) ^ V(O2));
	using EBitOr          = decltype(V(O1) | V(O2));
	using EAnd            = decltype(V(O1) && V(O2));
	using EOr             = decltype(V(O1) || V(O2));

	using EAssign         = decltype(V(O1) = V(O2));
	using EAssignAdd      = decltype(V(O1) += V(O2));
	using EAssignSub      = decltype(V(O1) -= V(O2));
	using EAssignMul      = decltype(V(O1) *= V(O2));
	using EAssignDiv      = decltype(V(O1) /= V(O2));
	using EAssignMod      = decltype(V(O1) %= V(O2));
	using EAssignLS       = decltype(V(O1) <<= V(O2));
	using EAssignRS       = decltype(V(O1) >>= V(O2));
	using EAssignAnd      = decltype(V(O1) &= V(O2));
	using EAssignXor      = decltype(V(O1) ^= V(O2));
	using EAssignOr       = decltype(V(O1) |= V(O2));

	using E1Assign1Add    = decltype(V(O1) = V(O1) + V(O2));
	using E1Assign2Add    = decltype(V(O1) = V(O1) + V(O2) + V(O3));
	using E2Assign1Add    = decltype(V(O1) = V(O2) = V(O3) + V(O3));

	using EIf             = decltype(If(V(O1), V(O2), V(O3)));
	using EIfAssign       = decltype(If(V(O1), V(O2), V(O3)) = V(O3));
	using EIfInAssign     = decltype(If(V(O1), V(O2) = V(O3), V(O3)));

	SECTION("ExpressionTagger") {
		SECTION("Unary operations") {
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EPlus>, Tags1R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EMinus>, Tags1R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ENot>, Tags1R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EBitNot>, Tags1R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EIndirection>, Tags1R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAddressOf>, Tags1R>);
		}

		SECTION("Binary operations") {
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EMultiplication>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EDivision>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EModulo>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAddition>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ESubtraction>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ELeftShift>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ERightShift>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ELower>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ELowerEqual>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EGreater>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EGreaterEqual>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EEqual>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<ENotEqual>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EBitAnd>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EBitXor>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EBitOr>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAnd>, Tags1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EOr>, Tags1R2R>);

			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssign>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignAdd>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignSub>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignMul>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignDiv>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignMod>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignLS>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignRS>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignAnd>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignXor>, Tags1W2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EAssignOr>, Tags1W2R>);
		}

		SECTION("If") {
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EIf>, Tags1R2R3R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EIfAssign>, Tags1R2W3W3R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<EIfInAssign>, Tags1R2W3R3R>);
		}

		SECTION("Multiple operations") {
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<E1Assign1Add>, Tags1W1R2R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<E1Assign2Add>, Tags1W1R2R3R>);
			TEST(REQUIRE, is_same_v<pfor::expr::ExpressionTagger<E2Assign1Add>, Tags1W2W3R3R>);
		}
	}
}
